;; @name fileio.seek
;; @desc
;;    Moves file pointer to given position in file.
;;    Distance to move is 64bit value.
;; @arg handle
;;    File handle (returned by {fileio.open}, {fileio.create} or {fileio.append}).
;; @arg distlow
;;    Low order dword of distance to move.
;; @arg disthigh
;;    High order dword of distance to move.
;; @arg mode
;;    Mode of seeking.
;;    0 = seek from beginning.
;;    1 = seek relative distance from current position.
;;    2 = seek from end (distance should be negative value).
;; @ret
;;    CF set on error
;; @note This procedure doesn't return new position in file.
;;       Use {fileio.tell} to get current position in file.
;; @note Seeking behind end of file is platform-dependant.
;;       On win32 it is allowed, but file doesn't increase until
;;       you write something to it.
;;       TODO - check this on linux.
;; @note This is platform-dependent procedure
;; @err ERR_NEGATIVE_SEEK
;;    Seek to negative file offset.
;; @err ERR_INVALID_HANDLE
;;    Invalid file handle.
;; @err ERR_OUT_OF_RANGE
;;    dist<0 when mode=0.
;; @err ERR_INVALID_MODE
;;    mode>2.
;; @err ERR_UNKNOWN
proc fileio.seek, handle, distlow, disthigh, mode
local high:DWORD
	push	ecx edx

	;check if mode is valid
	cmp	[mode], 2
	jbe	@f
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
@@:

	;check if dist is positive, if [.mode] is 0  (windoze would check this, but this is nicer)
	cmp	[mode], 0
	jne	@f
	cmp	[disthigh], 0
	jge	@f
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
@@:

	;call API
	mov	eax, [disthigh]
	mov	[high], eax
	push	[mode]			;windows uses same numbers as we do
	lea	eax, [high]
	push	eax			;ptr to high order dword of distance to move
	push	[distlow]
	push	[handle]
	call	[SetFilePointer]

	;check for error
	;we have to call GetLastError when -1 is returned
	cmp	eax, -1
	jne	@f
	call	[GetLastError]
	cmp	eax, 0
	jne	.error_SetFilePointer
	@@:

	mov	eax, -1 	;make sure noone uses return value

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_SetFilePointer:
	call	[GetLastError]

	;handle invalid type of handle :)
	push	ERR_INVALID_HANDLE
	cmp	eax, 6		;ERROR_INVALID_HANDLE
	je	@f
	cmp	eax, 132	;ERROR_SEEK_ON_DEVICE
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	;handle negative seek
	push	ERR_NEGATIVE_SEEK
	cmp	eax, 131	;ERROR_NEGATIVE_SEEK
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc
endp