;; @name mem.alloc
;; @desc
;;     Allocates block of memory.
;; @arg size
;;     Size of block to allocate, in bytes.
;; @ret
;;     CF set on error, otherwise
;;     EAX = pointer to allocated memory block
;; @err ERR_OUT_OF_MEM
;;     Not enough memory
;; @err ERR_ZERO_SIZE
;;     size = 0
;; @err ERR_OUT_OF_RANGE
;;     size is negative (or larger than 7FFFFFFFh if viewed as unsigned)
;; @err ERR_HEAP_CORRUPT
;;     {heap:Heap is corrupted}. You was writing to memory outside of allocated block.
;; @err ERR_MODULE_NOT_INITIALIZED
;;      Module is not initialized. Initialize module with {mem.init} to use this procedure.
;; @err ERR_UNKNOWN
;;     System returned error that wasn't translated to FASMLIB error.
;;     You can {contact} author about this, and translation can be added.
;; @dep mem.realloc_heap
;;     Only on platforms where FASMLIB heap manager is used.
;; @warn Don't forget to release allocated block with {mem.free}!
;; @warn Contents of allocated memory are unknown (use {mem.alloc0} to get block zeroed).
;; @note Returned pointer is always >= 10000h to make it easily distinguishable from handles.
;; @note Maximal allowed size is 7FFFFFFFh, to catch errors with negative values.
;; @note Altough returned block may have bigger size than requested on some implementations,
;;       don't use this additional memory. {mem.realloc} doesn't preserve this memory.
;; @note Memory management must be {mem.init:initialized} for this procedure to work.
;; @note This is platform-specific procedure
proc mem.alloc size
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	jz	.error_not_initialized

	;error if signed size<=0
	cmp	[size], 0
	je	.error_size_zero
	jl	.error_size_too_big

	;size += header
	mov	ecx, [size]
	add	ecx, 8

	;call HeapAlloc
	push	ecx
	push	dword 0 ;no flags
	push	[mem.heap]
	call	[HeapAlloc]
	or	eax,eax
	jz	.error_HeapAlloc

	;fill header
	mov	dword [eax], 'RnR'  ;sign
	mov	ecx, [size]
	mov	dword [eax+4], ecx  ;size (not including header)

	;return pointer behind header
	inc	[mem.alloc_count]
	add	eax, 8

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_HeapAlloc:
	call	[GetLastError]

	;catch out-of-memory
	push	ERR_OUT_OF_MEM
	cmp	eax, 8			;ERROR_NOT_ENOUGH_MEMORY
	je	@f
	cmp	eax, 14 		;ERROR_OUTOFMEMORY
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_size_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_size_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp