;=============================================================================
;; @name mem.realloc
;; @desc
;;    Rellocates (resizes) block of memory returned from {mem.alloc}, {mem.alloc0},
;;    {mem.realloc}, or {mem.realloc0}.
;;    Block can be moved during this process, so you need to update pointer(s) to it.
;;    Contents of block are preserved.
;;    If block is enlarging, contents of added space are unknown, use {mem.realloc0} to have them zeroed.
;; @arg blockptr
;;    Pointer to block to relocate.
;;    This must be same pointer as returned from allocation function, pointer into block is not enough.
;; @arg newsize
;;    New size of block, in bytes
;; @ret
;;    CF set on error, orherwise
;;    EAX = pointer to reallocated memory block
;; @err ERR_INVALID_POINTER
;;    Blockptr is not a valid allocated block pointer.
;;    This could also mean {heap:heap corruption}, these two cases cannot be properly distinguished.
;; @err ERR_OUT_OF_MEM
;;     Not enough memory
;; @err ERR_ZERO_SIZE
;;     size = 0
;; @err ERR_OUT_OF_RANGE
;;     size is negative (or larger than 7FFFFFFFh if viewed as unsigned)
;; @err ERR_HEAP_CORRUPT
;;     {heap:Heap is corrupted}. You was writing to memory outside of allocated block.
;; @err ERR_MODULE_NOT_INITIALIZED
;;      Module is not initialized. Initialize module with {mem.init} to use this procedure.
;; @err ERR_UNKNOWN
;;     System returned error that wasn't translated to FASMLIB error.
;;     You can {contact} author about this, and translation can be added.
;; @dep mem.realloc_heap
;;     Only on platforms where FASMLIB heap manager is used.
;; @warn
;;     Block can be moved during realocation! If you had any pointers to block,
;;     they become invalid (so rather prefer using indexes to allocated blocks).
;;     If you saved block pointer in variable, you need to rewrite it with
;;     returned value.
;; @note
;;     If block was enlarging, contents of added memory are unknown.
;;     Use {mem.realloc0} to have them zeroed.
;; @note Returned pointer is always >= 10000h to make it easily distinguishable from handles.
;; @note Maximal allowed newsize is 7FFFFFFFh, to catch errors with negative values.
;; @note Altough returned block may have bigger size than requested on some implementations,
;;       don't use this additional memory. {mem.realloc} doesn't preserve this memory.
;; @note Memory management must be {mem.init:initialized} for this procedure to work.
;; @note This is platform-specific procedure
proc mem.realloc blockptr, newsize
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	jz	.error_not_initialized

	;error if size=0
	cmp	[newsize], 0
	je	.error_newsize_zero
	jl	.error_newsize_too_big

	;EAX = memory block - header
	mov	eax, [blockptr]
	sub	eax, 8

	;check if it is our memory block
	cmp	dword [eax], 'RnR'
	je	@f
	mov	eax, ERR_INVALID_POINTER
	jmp	.rc
@@:

	;call HeapReAlloc
	mov	ecx, [newsize]
	add	ecx, 8			;size of block + header
	push	ecx
	push	eax			;pointer to old memory block before header
	push	0			;no flags
	push	[mem.heap]		;our heap handle
	call	[HeapReAlloc]
	or	eax,eax
	jz	.error_HeapRealloc

	;fill new header (needed when block has moved)
	mov	dword [eax], 'RnR'	;sign
	mov	ecx, [newsize]
	mov	[eax+4], ecx		;size (not including header)

	;return pointer behind header
	add	eax, 8

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_HeapRealloc:
	call [GetLastError]

	;catch out-of-memory
	push	ERR_OUT_OF_MEM
	cmp	eax, 8			;ERROR_NOT_ENOUGH_MEMORY
	je	@f
	cmp	eax, 14 		;ERROR_OUTOFMEMORY
	je	@f
	mov	[esp],eax
@@:	pop	eax
	je	.rc

	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_newsize_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_newsize_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp