;; @name mem.uninit
;; @desc
;;     Unitializes FASMLIB memory managent.
;;     When module is not initialized, you cannot use following procedures:
;;     {mem.alloc}, {mem.alloc0}, {mem.realloc}, {mem.realloc0}, {mem.free}, {mem.size}.
;;     This procedure also checks for {heap:heap corruption} and {memory leak:memory leaks}.
;; @ret
;;     CF set on error
;; @err ERR_MEMORY_LEAK
;;     You've got a {memory leak}. Check all places where you call {mem.alloc}
;;     or {mem.alloc0}, whether you always release allocated block with {mem.free}.
;;     Also check if you didn't forget to {stdstream.uninit:uninitialize standard streams}
;;     prior to calling this procedure.
;; @err ERR_HEAP_CORRUPT
;;     {heap:Heap is corrupted}. You was writing to memory outside of allocated block.
;; @err ERR_MODULE_NOT_INITIALIZED
;;      Module is not initialized. Initialize module with {mem.init} first.
;; @err ERR_UNKNOWN
;;     System returned error that wasn't translated to FASMLIB error.
;;     You can {contact} author about this, and translation can be added.
;; @dep mem.free_heap
;;     Only on platforms where FASMLIB heap manager is used
;; @note This procedure is called by {fasmlib.uninit}.
;;       You can use it for lazy uninitialization of entire library.
;; @note This is platform-specific procedure
proc mem.uninit
	push	ecx edx

	;is module initialized?
	cmp	[mem.initialized],0
	jz	.error_not_initialized

	;check memory leak
	cmp	[mem.alloc_count],0
	ja	.error_memory_leak

	;store state
	mov	[mem.initialized],0

.rnc:	clc
.r:	pop	edx ecx
	ret
.rc:	stc
	jmp	.r

.error_HeapDestroy:
	mov	eax, ERR_UNKNOWN
	jmp	.rc

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_memory_leak:
	mov	eax, ERR_MEMORY_LEAK
	jmp	.rc
endp