;; @name extconv.d_dec
;; @desc
;;    Converts unsigned dword (32bit) value to decimal string.
;;    Supports padding of output with spaces or zeroes.
;;    For simpler version see {conv.d_dec}.
;; @arg dest
;;    Pointer to destination string buffer.
;; @arg destlen
;;    Size of destination string buffer.
;; @arg value
;;    Dword value to convert.
;; @arg padding
;;    Type of padding:
;;    type 0 is no padding,
;;    type 1 is prepending with spaces,
;;    type 2 is prepending with leading zeroes.
;; @arg padlen
;;    Length to which should output be padded.
;;    If output is already longer than this value before padding, then output is not padded.
;;    Ignored if padding = 0.
;; @arg len
;;    Pointer to dword variable, which will on return hold length of output (terminating zero not included).
;;    Ignored if zero.
;; @ret
;;    CF=1 on error.
;; @err ERR_BUFFER_FULL
;;    Destination buffer is not big enough.
;;    Only first destlen bytes of resulting string were written.
;;    String in destination buffer is NOT zero terminated.
;;    Variable pointed by len contains destlen
;; @err ERR_ZERO_SIZE
;;    destlen=0.
;;    Variable pointed by len is set to 0.
;; @err ERR_INVALID_MODE
;;    padding > 2.
;;    Variable pointed by len is set to 0.
proc extconv.d_dec dest, destlen, value, padding, padlen, len
	push	ebx ecx edx edi

	cmp	[destlen], 0
	je	untaken .error_destlen_zero
	cmp	[padding], 2
	ja	untaken .error_padding_type

	mov	edi, [dest]

	;convert value (push remainders to stack as dwords)
	xor	ecx, ecx		;ecx = number of chars
	mov	eax, [value]		;eax = value
	mov	ebx, 10
.push_digit:
	xor	edx, edx
	div	ebx
	push	edx
	inc	ecx
	cmp	eax, 0
	jnz	taken .push_digit


	mov	ebx, [destlen]

	;get padding type
	cmp	[padding], 0
	je	.padded
	mov	dl, ' '
	cmp	[padding], 1
	je	@f
	mov	dl, '0'
	@@:

	;pad up to padlen
	mov	eax, [padlen]
.pad:	cmp	eax, ecx
	jbe	.padded
	mov	byte [edi], dl
	inc	edi
	dec	ebx
	jz	untaken .error_buffer_full
	dec	eax
	jmp	.pad
.padded:

	;pop values from stack and store them
	;EDI = dest, EBX=destlen
.store_digit:
	pop	eax
	mov	al, [conv.hextable + eax]
	mov	[edi], al
	inc	edi
	dec	ebx
	jz	.error_buffer_full
	dec	ecx
	jnz	taken .store_digit

	;store ending zero
	mov	byte [edi], 0

	;save len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	sub	edi, [dest]
	mov	[eax], edi
	@@:

	clc

.r:	pop	edi edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_buffer_full:
	;restore stack
	lea	esp, [esp + 4*(ecx-1)]

	;set len to destlen
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	push	[destlen]
	pop	dword [eax]
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_padding_type:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp