;; @name extconv.hex_q
;; @desc
;;    Converts hexadecimal string to unsigned qword (64bit) value.
;;    This is extended version, for simpler version see {conv.hex_q}.
;; @arg string
;;    Pointer to string to convert.
;; @arg buflen
;;    Size of string's buffer.
;;    You can set this to -1 (FFFFFFFFh) if string is quaranteed to be properly zero-terminated.
;; @arg len
;;    Pointer to dword variable, which will on return hold number of converted characters.
;;    Ignored if zero.
;; @ret
;;    CF=1 on error, otherwise
;;    EDX:EAX = value.
;; @err ERR_NOT_HEXADECIMAL
;;    First character of string is not hexadecimal digit.
;; @err ERR_OUT_OF_RANGE
;;    Value is > FFFFFFFFFFFFFFFF
;; @err ERR_ZERO_SIZE
;;    strlen=0
;; @note This procedure converts digits until it reaches non-digit character,
;;       or until value overflows.
;; @nodeps
proc extconv.hex_q string, strlen, len
	push	esi ecx ebx

	;error is strlen=0
	cmp	[strlen], 0
	je	.error_zero_size

	;ESI = string
	mov	esi, [string]

	;error if first char isn't digit
	cmp	byte [esi], '0'
	jb	.error_not_hex
	cmp	byte [esi], '9'
	jbe	@f
	cmp	byte [esi], 'A'
	jb	.error_not_hex
	cmp	byte [esi], 'F'
	jbe	@f
	cmp	byte [esi], 'a'
	jb	.error_not_hex
	cmp	byte [esi], 'f'
	ja	.error_not_hex
@@:

	;EDX:EBX = final value, for now 0
	mov	edx, 0
	mov	ebx, 0

	;ECX = count of converted chars
	mov	ecx, 0

	;loop
.digit_loop:
	;AL=char
	mov	al, [esi]
	inc	esi

	;end if it isn't hex digit
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	jbe	.numeric
	cmp	al, 'A'
	jb	.done
	cmp	al, 'F'
	jbe	.upcase_letter
	cmp	al, 'a'
	jb	.done
	cmp	al, 'f'
	ja	.done

.lowcase_letter:
	sub	al, 'a'-10
	jmp	.add_digit

.upcase_letter:
	sub	al, 'A'-10
	jmp	.add_digit

.numeric:
	sub	al, '0'

.add_digit:
	;test overflow
	test	edx, 0xF0000000
	jnz	.overflow

	;add digit
	shld	edx, ebx, 4
	shl	ebx, 4
	movzx	eax, al
	add	ebx, eax
	adc	edx, 0

	;continue loop
	inc	ecx
	cmp	ecx, [strlen]
	jb	.digit_loop
.done:

	;fill len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;return EDX:EBX
	mov	eax, ebx

	clc
.r:	pop	ebx ecx esi
	ret
.rc:	stc
	jmp	.r


.overflow:
	;set len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], ecx
	@@:

	;return error
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_zero_size:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc


.error_not_hex:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	;return error
	mov	eax, ERR_NOT_HEXADECIMAL
	jmp	.rc
endp