;; @name extconv.iq_dec
;; @desc
;;    Converts signed qword (64bit) value to decimal string.
;;    Supports padding of output with spaces or zeroes, and explicit plus sign.
;;    For simpler version see {conv.iq_dec}.
;; @arg dest
;;    Pointer to destination string buffer.
;; @arg destlen
;;    Size of destination string buffer.
;; @arg valuelow
;;    Low order dword of value to convert.
;; @arg valuehigh
;;    High order dword of value to convert.
;; @arg padding
;;    Type of padding:
;;    Type 0 is no padding.
;;    Type 1 is prepending with spaces, and sign counts into padding size.
;;    Type 2 is prepending with leading zeroes. In this type padlen specifies number
;;    digits displayed. One more character is always printed as sign (' ', '+' or '-').
;; @arg padlen
;;    Length to which should output be padded.
;;    If output is already longer than this value before padding, then output is not padded.
;;    Ignored if padding = 0.
;; @arg plus
;;    Set to 1 if you want to display plus sign on positive numbers.
;; @arg len
;;    Pointer to dword variable, which will on return hold length of output (terminating zero not included).
;;    Ignored if zero.
;; @ret
;;    CF=1 on error.
;; @err ERR_BUFFER_FULL
;;    Destination buffer is not big enough.
;;    Only first destlen bytes of resulting string were written.
;;    String in destination buffer is NOT zero terminated.
;;    Variable pointed by len contains destlen
;; @err ERR_ZERO_SIZE
;;    destlen=0.
;;    Variable pointed by len is set to 0.
;; @err ERR_INVALID_MODE
;;    padding > 2, or plus > 1
;;    Variable pointed by len is set to 0.
;; @nodeps
proc extconv.iq_dec dest, destlen, valuelow, valuehigh, padding, padlen, plus, len
	push	ebx ecx edx edi esi

	cmp	[destlen], 0
	je	.error_destlen_zero
	cmp	[padding], 2
	ja	.error_padding_type
	cmp	[plus], 1
	ja	.error_plus

	;EDI:EAX = abs(value)
	mov	edi, [valuehigh]
	mov	eax, [valuelow]
	cmp	edi, 0
	jge	@f
	not	edi		;neg EDI:EAX
	not	eax
	add	eax, 1
	adc	edi, 0
	@@:

	;convert value (push remainders to stack as dwords)
	xor	ecx, ecx		;ecx = number of chars
	mov	ebx, 10
.push_digit:
	xor	edx, edx
	xchg	eax, edi
	div	ebx
	xchg	edi, eax
	div	ebx
	push	edx
	inc	ecx
	cmp	edi, 0
	jnz	.push_digit
	cmp	eax, 0
	jnz	taken .push_digit

	mov	edi, [dest]
	mov	ebx, [destlen]

	;skip padding if not wanted
	cmp	[padding], 0
	je	.padding_done

	;ESI = number of characters to be padded
	mov	esi, ecx

	;decide padding type
	cmp	[padding], 1
	je	.padding_with_space

.padding_with_zero:
	mov	dl, '0'

	;we need another character for padding
	dec	ebx
	jz	.error_buffer_full

	;if value is negative, print '-'
	cmp	[valuehigh], 0
	jge	@f
	mov	byte [edi], '-'
	inc	edi
	jmp	.do_padding
	@@:

	;value is positive, if plus=0, print space
	cmp	[plus], 0
	jne	@f
	mov	byte [edi], ' '
	inc	edi
	jmp	.do_padding
	@@:

	;value is positive, plus=1, print '+' if plus=1
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '+'
	inc	edi

	jmp	.do_padding


.padding_with_space:
	mov	dl, ' '

	;if number is negative, we need to pad one character less
	;to reserve place for '-' sign
	cmp	[valuehigh], 0
	jge	@f
	inc	esi
	jmp	.do_padding
	@@:

	;number is positive, if plus=1, we need to reserve place for '+'
	cmp	[plus], 1
	jne	@f
	inc	esi
	@@:

	;do padding
	;EDI = dest, EBX=destlen, ESI=number of characters to pad, DL=padding char
.do_padding:

	;pad ESI up to padlen
	mov	eax, [padlen]
.pad:	cmp	eax, esi
	jbe	.padding_done
	mov	byte [edi], dl
	inc	edi
	dec	ebx
	jz	.error_buffer_full
	dec	eax
	jmp	.pad
.padding_done:

.print_sign:

	;if padding type=2, we already have printed sign before padding
	cmp	[padding], 2
	je	.sign_done

	;if value is negative, print '-'
	cmp	[valuehigh], 0
	jge	@f
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '-'
	inc	edi
	jmp	.sign_done
	@@:

	;value is positive, print '+' if plus=1
	cmp	[plus], 1
	jne	.sign_done
	dec	ebx
	jz	.error_buffer_full
	mov	byte [edi], '+'
	inc	edi

.sign_done:

	;pop values from stack and store them
	;EDI = dest, EBX=destlen, ECX=number of digits on stack
.store_digit:
	pop	eax
	mov	al, [conv.hextable + eax]
	mov	[edi], al
	inc	edi
	dec	ebx
	jz	.error_buffer_full
	dec	ecx
	jnz	.store_digit

	;store ending zero
	mov	byte [edi], 0

	;save len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	sub	edi, [dest]
	mov	[eax], edi
	@@:

	clc

.r:	pop	esi edi edx ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_buffer_full:
	;restore stack
	lea	esp, [esp + 4*(ecx-1)]

	;set len to destlen
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	push	[destlen]
	pop	dword [eax]
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_destlen_zero:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_plus:
.error_padding_type:
	;set len to 0
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp
