;; @name file.extappend
;; @desc
;;    Opens existing file for writing, and sets pointer to end of file.
;; @arg objbuf
;;    Pointer to buffer for {fstream:FSTREAM} object.
;;    If 0, object is allocated.
;; @arg name
;;    Name of file to create
;; @arg create
;;    Specifies whether to create file if it doesn't exist.
;;    1=create, 0=don't create.
;; @arg mode
;;    Access mode to file: 1=write only, 2=read/write.
;; @arg wbuf
;;    RESERVED, must be 0
;; @arg wbuflen
;;    RESERVED, must be 0
;; @arg rbuf
;;    Pointer to read buffer.
;;    If 1, buffer will be allocated, and wbuflen holds initial size.
;;    If 0, read buffering is disabled.
;;    Must be 0 if mode = 1.
;; @arg rbuflen
;;    If rbuf=pointer, this is size of buffer.
;;    If rbuf=1, this is initial size of buffer.
;;    If rbuf=0, ignored.
;;    Must be 0 if mode = 1.
;; @ret
;;    CF = 1 on error, otherwise
;;    EAX = stream handle
;; @err ERR_FILE_NOT_FOUND
;;    File with given name wasn't found.
;;    Only returned when create=0.
;; @err ERR_CANT_OPEN_FILE
;;    File can't be created.
;; @err ERR_INVALID_MODE
;;    mode = 0, or mode > 2, or overwr > 1,
;;    or rbuf is nonzero when mode=1, or rbuflen is nonzero when mode=1.
;; @err ERR_UNKNOWN
;;    Platform-specific error during creating file.
;; @warn Don't forget to close returned file stream with {stream.close}!
;; @note This is extended version, see {file.append} for simpler usage.
;; @note Opens file as stream, if you want direct file I/O use {fileio.append}.
proc file.extappend objbuf, name, create, mode, wbuf, wbuflen, rbuf, rbuflen
	push	ebx edi

	;check if mode is 1 or 2
	mov	eax, [mode]
	test	eax, eax
	jz	.error_invalid_mode
	cmp	eax, 2
	ja	.error_invalid_mode

	;check if read buffer is disabled when mode=1
	cmp	eax, 1
	jne	@f
	cmp	[rbuf], 0
	jne	.error_invalid_mode
	cmp	[rbuflen], 0
	jne	.error_invalid_mode
	@@:

	;check if create is 0 or 1
	cmp	[create], 2
	jae	.error_invalid_mode

	;EBX = file handle
	libcall fileio.append, [name], [create], [mode]
	jc	.rc
	mov	ebx, eax

	;EDI = stream object
	libcall stream.create,\
		[objbuf],\			      ;pass info from caller
		sizeof.FSTREAM - sizeof.STREAM,\      ;size of additional data
		fstream._vtab,\ 		      ;virtual table
		[rbuf], [rbuflen],\		      ;pass read buffer info from caller
		[wbuf], [wbuflen]		      ;pass write buffer info from caller
	jc	.rc2
	mov	edi, eax

	;store additional data
	mov	[edi+FSTREAM.handle], ebx
	mov	eax, [mode]
	mov	[edi+FSTREAM.mode], eax
	mov	[edi+FSTREAM.closeh], 1

	;return stream object/handle
	mov	eax, edi

.rnc:	clc
.r:	pop	edi ebx
	ret
.rc:	stc
	jmp	.r

;close file handle in EBX and return error
.rc2:
	push	eax
	libcall fileio.close, ebx
	pop	eax
	jmp	.rc


.error_invalid_mode:
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp