;; @name file.extopen
;; @desc
;;    Opens existing file as stream.
;; @arg objbuf
;;    Pointer to buffer for {fstream:FSTREAM} object.
;;    If 0, object is allocated.
;; @arg name
;;    Pointer to name of file to open. Must be zero-terminated string.
;; @arg mode
;;    Access mode to file: 0=read only, 1=write only, 2=read/write.
;; @arg wbuf
;;    RESERVED, must be 0
;; @arg wbuflen
;;    RESERVED, must be 0
;; @arg rbuf
;;    Pointer to read buffer.
;;    If 1, buffer will be allocated, and wbuflen holds initial size.
;;    If 0, read buffering is disabled.
;; @arg rbuflen
;;    If rbuf=pointer, this is size of buffer.
;;    If rbuf=1, this is initial size of buffer.
;;    If rbuf=0, ignored.
;; @ret
;;    CF = 1 on error, otherwise
;;    EAX = stream handle
;; @err ERR_FILE_NOT_FOUND
;;    File with specified name wasn't found.
;; @err ERR_CANT_OPEN_FILE
;;    File exists, but couldn't be opened.
;; @err ERR_INVALID_MODE
;;    mode > 2.
;; @err ERR_UNKNOWN
;;    Platform-specific error during opening file.
;; @warn Don't forget to close returned file stream with {stream.close}!
;; @warn If you disable read buffering (rbuf=0), then you can't use {stream.peek},
;;       or any procedures that require peek.
;;       Most {text read:text reading} procedures will fail.
;; @note This is extended version, see {file.open} for simpler usage.
;; @note Opens file as stream, if you want direct file I/O use {fileio.open}.
proc file.extopen objbuf, name, mode, wbuf, wbuflen, rbuf, rbuflen
	push	ebx edi

	;error if mode > 2
	cmp	[mode], 2
	ja	.error_invalid_mode

	;EBX = file handle
	libcall fileio.open, [name], [mode]
	jc	.rc
	mov	ebx, eax

	;EDI = stream object
	libcall stream.create,\
		[objbuf],\				;pass info from caller
		sizeof.FSTREAM - sizeof.STREAM,\	;size of additional data
		fstream._vtab,\ 			;virtual table
		[rbuf], [rbuflen],\			;pass read buffer info from caller
		[wbuf], [wbuflen]			;pass write buffer info from caller
	jc	.error_close_ebx
	mov	edi, eax

	;store additional data
	mov	[edi+FSTREAM.handle], ebx
	mov	eax, [mode]
	mov	[edi+FSTREAM.mode], eax
	mov	[edi+FSTREAM.closeh], 1

	;return stream handle
	mov	eax, edi

.rnc:	clc
.r:	pop	edi ebx
	ret
.rc:	stc
	jmp	.r

.error_close_ebx:
	push	eax
	libcall fileio.close, ebx
	pop	eax
	jmp	.rc

.error_invalid_mode:
	mov	eax, ERR_INVALID_MODE
	jmp	.rc
endp