;============================================================================
; mem.alloc
; desc: allocates <size> bytes of memory
; args: size - size of memory to allocate
; ret: CF set on error, otherwise
;      eax = pointer to beginning of allocated memory block
; note: - returned pointer must be always >= 10000h
;       - contents of allocated memory are unknown
;         (use alloc0 to have block zeroed)
;       - don't forget to mem.free allocated memory!
;       - maximal allowed size if 7FFFFFFFh, to catch errors with negative value
; error: ERR_OUT_OF_MEM   - out of memory
;        ERR_ZERO_SIZE    - size=0
;        ERR_OUT_OF_RANGE - size >= 80000000h
;        ERR_HEAP_CORRUPT - you was overwriting memory outside allocated block
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;        mem.heap_alloc
;============================================================================
proc mem.alloc size
	push	ebx ecx edx esi edi

	;is module initialized?
	cmp	[mem.initialized], 0
	je	.error_not_initialized

	;was heap corruption detected?
	cmp	[mem.heap_corrupt], 0
	jne	.error_heap_corrupt

	;error if size=0
	cmp	[size], 0
	je	.error_size_zero
	jl	.error_size_too_big

	;ECX = size to allocate, aligned to HEAP_BLOCK_ALIGN
	mov	ecx, [size]
	add	ecx, HEAP_BLOCK_ALIGN-1
	and	ecx, not (HEAP_BLOCK_ALIGN-1)

	;EBX = pointer to first HEAP block
	mov	ebx, [mem.heap.start]

	;find unused block with enough space
.find_block:

	;check if block is okay
	cmp	[ebx + HEAPBLOCK.sign], 'HB'
	jne	.error_heap_corrupt

	;is the block unused?
	cmp	[ebx + HEAPBLOCK.used], 1
	ja	.error_heap_corrupt
	je	.try_next_block

	;is the block large enough?
	cmp	[ebx + HEAPBLOCK.size], ecx
	jae	.block_found

.try_next_block:

	;is there some next block?
	cmp	[ebx + HEAPBLOCK.last], 1
	ja	.error_heap_corrupt
	je	.no_block_found

	;go to next block
	add	ebx, [ebx + HEAPBLOCK.size]
	add	ebx, sizeof.HEAPBLOCK
	jmp	.find_block

	;no suitable block found. we need to enlarge heap
	;EBX = pointer to last block
.no_block_found:
.resize_heap:

	; increasize heap size by needed size + header size aligned to HEAP_SIZE_ALIGN
	mov	edx, ecx
	add	edx, sizeof.HEAPBLOCK
	add	edx, HEAP_SIZE_ALIGN-1
	and	edx, not (HEAP_SIZE_ALIGN-1)
	add	edx, [mem.heap.size]
	libcall mem.realloc_heap, [mem.heap.start], edx, [mem.heap.size]
	mov	[mem.heap_corrupt], 1  ;if this call failed, we are screwed
	jc	.rc
	mov	[mem.heap_corrupt], 0
	xchg	[mem.heap.size], edx

	;edx = size of added block
	mov	eax, [mem.heap.size]
	sub	eax, edx
	mov	edx, eax

	;if last block was free, we append new space to it
	cmp	[ebx + HEAPBLOCK.used], 1
	jne	.add_new_space_to_last_block

	;last block is used, but first we need to check if we can take enough unused space from last block
	;if there is at least sizeof.HEAPBLOCK unused bytes after alignation, we take them
	mov	esi, [ebx + HEAPBLOCK.asize]
	add	esi, HEAP_BLOCK_ALIGN-1
	and	esi, not (HEAP_BLOCK_ALIGN-1)
	mov	eax, [ebx + HEAPBLOCK.size]
	sub	eax, esi
	jc	.create_block_in_added_space  ;if size is not even aligned. shouldn't happen but anyway...

	cmp	eax, sizeof.HEAPBLOCK
	jb	.create_block_in_added_space

	;EBX = last block, ESI = new size of EBX block
.take_space_from_last_block:

	;set up EBX
	mov	[ebx + HEAPBLOCK.last], 0
	mov	[ebx + HEAPBLOCK.size], esi

	;set up ESI
	lea	esi, [ebx + esi + sizeof.HEAPBLOCK]
	mov	[esi + HEAPBLOCK.sign], 'HB'
	mov	[esi + HEAPBLOCK.last], 1
	mov	[esi + HEAPBLOCK.prev], ebx
	mov	eax, [mem.heap.start]
	add	eax, [mem.heap.size]
	sub	eax, esi
	sub	eax, sizeof.HEAPBLOCK
	mov	[esi + HEAPBLOCK.size], eax  ;size = up to end of block
	mov	[esi + HEAPBLOCK.used], 0

	mov	ebx, esi
	jmp	.block_found

.create_block_in_added_space:

	;esi = new block
	lea	esi, [ebx + sizeof.HEAPBLOCK]
	add	esi, [ebx + HEAPBLOCK.size]

	;set up the block
	mov	[ebx + HEAPBLOCK.last], 0
	mov	[esi + HEAPBLOCK.sign], 'HB'
	mov	[esi + HEAPBLOCK.last], 1
	mov	[esi + HEAPBLOCK.prev], ebx
	sub	edx, sizeof.HEAPBLOCK
	mov	[esi + HEAPBLOCK.size], edx
	mov	[esi + HEAPBLOCK.used], 0

	mov	ebx, esi
	jmp	.block_found


.add_new_space_to_last_block:
	add	[ebx + HEAPBLOCK.size], edx

	;suitable block found, we will use it for allocation
	;EBX = pointer to block, ecx=size to allocate
.block_found:

	;EDI = beginning of new block aligned
	lea	edi, [ebx + sizeof.HEAPBLOCK + ecx]

	;EDX = space left in block
	mov	edx, [ebx + HEAPBLOCK.size]
	sub	edx, ecx
	jz	.set_block ;if zero, we can skip it

	;ESI = next block
	cmp	[ebx + HEAPBLOCK.last], 1
	ja	.error_heap_corrupt
	je	.is_last_block
	lea	esi, [ebx + sizeof.HEAPBLOCK]
	add	esi, [ebx + HEAPBLOCK.size]

	jmp	.not_last_block


.is_last_block:
	sub	edx, sizeof.HEAPBLOCK

	;if following space is not big enough, leave it as padding
	cmp	edx, HEAP_BLOCK_ALIGN
	jb	.set_block

	;there is enough space, create new block (EDI)
	mov	[ebx + HEAPBLOCK.last], 0
	mov	[ebx + HEAPBLOCK.size], ecx
	mov	[edi + HEAPBLOCK.sign], 'HB'
	mov	[edi + HEAPBLOCK.used], 0
	mov	[edi + HEAPBLOCK.last], 1
	mov	[edi + HEAPBLOCK.size], edx
	mov	[edi + HEAPBLOCK.prev], ebx

	jmp	.set_block

.not_last_block:

	;if next block is unused, move it to rest of buffer append space
	cmp	[esi + HEAPBLOCK.used], 0
	jne	.next_block_used

.next_block_unused:

	;no moving if gained space is <= sizeof.HEAPBLOCK (could case problems)
	cmp	edx, sizeof.HEAPBLOCK
	jbe	.set_block

	;next block is unused, move it to EDI
	mov	[ebx + HEAPBLOCK.size], ecx
	mov	[esi + HEAPBLOCK.sign], 0 ;clean sign of the next block, to prevent identifying it as valid block in future
	mov	[edi + HEAPBLOCK.sign], 'HB'
	mov	[edi + HEAPBLOCK.used], 0     ;newblock.used = 0
	mov	al, [esi + HEAPBLOCK.last]
	mov	[edi + HEAPBLOCK.last], al
	mov	eax, [esi + HEAPBLOCK.size]
	add	eax, [edx - sizeof.HEAPBLOCK]
	add	[edi + HEAPBLOCK.size], eax   ;newblock.size = newblock.size + nextblock.size - nextblock_header_size
	mov	[edi + HEAPBLOCK.prev], ebx   ;newblock.prev = previous block (EBX)
	mov	[eax + HEAPBLOCK.prev], edi   ;nextblock.next.prev = newblock

	jmp	.set_block

	;following block (ESI) is used
.next_block_used:
	sub	edx, sizeof.HEAPBLOCK

	;if remaining space is not big enough to create new block, leave it as padding
	cmp	edx, HEAP_BLOCK_ALIGN
	jb	.set_block

	;it's big enough, so create new block
	mov	[ebx + HEAPBLOCK.size], ecx
	mov	[edi + HEAPBLOCK.sign], 'HB'
	mov	[edi + HEAPBLOCK.size], edx
	mov	[edi + HEAPBLOCK.prev], ebx
	mov	[edi + HEAPBLOCK.used], 0
	mov	[edi + HEAPBLOCK.last], 0
	lea	eax, [edi + edx + sizeof.HEAPBLOCK]
	mov	[eax + HEAPBLOCK.prev], edi		;ebx->next->prev = ebx->next



	;set current block we are going to allocate
.set_block:
	mov	eax, [size]
	mov	[ebx + HEAPBLOCK.asize], eax
	mov	[ebx + HEAPBLOCK.used], 1

	;another block successfully allocated
	inc	[mem.alloc_count]

	;return pointer to block
	lea	eax, [ebx + sizeof.HEAPBLOCK]

.rnc:	clc
.r:	pop	esi edi edx ecx ebx
	ret
.rc:	stc
	ret

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_heap_corrupt:
	mov	[mem.heap_corrupt], 1
	mov	eax, ERR_HEAP_CORRUPT
	jmp	.rc

.error_size_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_size_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

endp