;=============================================================================
; mem.free
; desc: deallocates memory block
; args: blockptr - pointer to memory block returned by mem.alloc
; ret: CF set on error
; note: after error, consider block as freed, don't use it anymore
; error: ERR_INVALID_POINTER - if blockptr isn't valid block pointer
;        ERR_HEAP_CORRUPT - you was overwriting memory outside allocated block
;        ERR_UNKNOWN
;        ERR_MODULE_NOT_INITIALIZED
;        ERR_FASMLIB
;=============================================================================
proc mem.free blockptr

	push	ebx edi

	;is module initialized?
	cmp	[mem.initialized], 0
	je	.error_not_initialized

	;was heap corruption detected?
	cmp	[mem.heap_corrupt], 0
	jne	.error_heap_corrupt

	;EBX = pointer to heap block
	mov	ebx, [blockptr]
	sub	ebx, sizeof.HEAPBLOCK

	;check if pointer lies inside heap
	mov	eax, ebx
	sub	eax, [mem.heap.start]
	jb	.error_bad_pointer
	sub	eax, [mem.heap.size]
	jae	.error_bad_pointer

	;check block
	cmp	[ebx + HEAPBLOCK.sign], 'HB'
	jne	.error_bad_pointer
	cmp	[ebx + HEAPBLOCK.used], 1
	ja	.error_heap_corrupt
	jb	.error_bad_pointer

	;free block
	mov	[ebx + HEAPBLOCK.used], 0

.check_following_block:

	;EDI = following block (if any)
	cmp	[ebx + HEAPBLOCK.last], 1
	ja	.error_heap_corrupt
	je	.check_previous_block
	lea	edi, [ebx + sizeof.HEAPBLOCK]
	add	edi, [ebx + HEAPBLOCK.size]

	;check following block
	cmp	[edi + HEAPBLOCK.sign], 'HB'
	jne	.error_heap_corrupt
	cmp	[edi + HEAPBLOCK.prev], ebx
	jne	.error_heap_corrupt

	;if following block is unused, append it
	cmp	[edi + HEAPBLOCK.used], 1
	ja	.error_heap_corrupt
	je	.check_previous_block

	;append block
	mov	[edi + HEAPBLOCK.sign], 0
	mov	eax, [edi + HEAPBLOCK.size]
	add	eax, sizeof.HEAPBLOCK
	add	[ebx + HEAPBLOCK.size], eax
	mov	al, [edi + HEAPBLOCK.last]
	mov	[ebx + HEAPBLOCK.last], al

.check_previous_block:

	;EDI = previous block (if any)
	mov	edi,[ebx + HEAPBLOCK.prev]
	test	edi, edi
	jz	.done

	;check EDI block
	cmp	[edi + HEAPBLOCK.sign], 'HB'
	jne	.error_heap_corrupt
	cmp	[edi + HEAPBLOCK.last], 0
	jne	.error_heap_corrupt
	lea	eax, [edi + sizeof.HEAPBLOCK]
	add	eax, [edi + HEAPBLOCK.size]
	cmp	eax, ebx
	jne	.error_heap_corrupt

	;if EDI block is unused, append EBX block to to it
	cmp	[edi + HEAPBLOCK.used], 1
	ja	.error_heap_corrupt
	je	.done

	;append block
	mov	[ebx + HEAPBLOCK.sign], 0
	mov	eax, [ebx + HEAPBLOCK.size]
	add	eax, sizeof.HEAPBLOCK
	add	[edi + HEAPBLOCK.size], eax
	mov	al, [ebx + HEAPBLOCK.last]
	mov	[edi + HEAPBLOCK.last], al

	mov	ebx, edi

.done:

	;if this isn't last block, set prev of next block
	cmp	[ebx + HEAPBLOCK.last], 1
	je	@f
	lea	eax, [ebx + sizeof.HEAPBLOCK]
	add	eax, [ebx + HEAPBLOCK.size]
	mov	[eax + HEAPBLOCK.prev], ebx
	@@:

	;another block successfully released
	dec	[mem.alloc_count]
	jc	.error_alloc_count_overflow

	;TODO - reduce heap size if ebx is last?

.rnc:	clc
.r:	pop	edi ebx
	ret
.rc:	stc
	ret

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_heap_corrupt:
	mov	[mem.heap_corrupt], 1
	mov	eax, ERR_HEAP_CORRUPT
	jmp	.rc

.error_bad_pointer:
	mov	eax, ERR_INVALID_POINTER
	jmp	.rc

.error_alloc_count_overflow:
	mov	eax, ERR_FASMLIB
	jmp	.rc
endp