;=============================================================================
; mem.realloc
; desc: reallocates memory block
; args: blockptr - pointer to memory block returned by mem.(re)alloc
;       newsize  - new size of memory block
; ret: CF set on error, orherwise
;      eax = pointer to reallocated memory block
; note: returned pointer must be always >= 10000h
;       - data in block are is modified (only data at the end of block is
;         lost if newsize < previous size)
;       - if block is enlarging, then contents of added memory are unknown
;         (use realloc0 to have it zeroed)
;       - block can move to other place in memory, so previous pointers into
;         block become invalid (so rather use indexes in block than pointers)
;       - maximal allowed size if 7FFFFFFFh, to catch errors with negative value
;       - if you had pointer stored in variable, don't forget to restore it:
;           libcall mem.realloc, [blockptr], [newsize]
;           jc error
;           mov [blockptr], eax     ;without this, it's hard-to-find bug
; error: ERR_INVALID_POINTER - blockptr isn't a valid block pointer
;        ERR_HEAP_CORRUPT    - you was overwriting memory outside allocated block
;        ERR_ZERO_SIZE       - newsize = 0
;        ERR_OUT_OF_RANGE    - newsize >= 80000000h
;        ERR_MODULE_NOT_INITIALIZED
;        +mem.alloc
;        +mem.free
;=============================================================================
proc mem.realloc blockptr, newsize
	push	ebx ecx edx esi edi

	;is module initialized?
	cmp	[mem.initialized], 0
	je	.error_not_initialized

	;was heap corruption detected?
	cmp	[mem.heap_corrupt], 0
	jne	.error_heap_corrupt

	;error if newsize=0
	cmp	[newsize], 0
	je	.error_newsize_zero
	jl	.error_newsize_too_big

	;EBX = pointer to heap block
	mov	ebx, [blockptr]
	sub	ebx, sizeof.HEAPBLOCK

	;check if pointer lies inside heap
	mov	eax, ebx
	sub	eax, [mem.heap.start]
	jb	.error_bad_pointer
	sub	eax, [mem.heap.size]
	jae	.error_bad_pointer

	;check block
	cmp	[ebx + HEAPBLOCK.sign], 'HB'
	jne	.error_bad_pointer
	cmp	[ebx + HEAPBLOCK.used], 1
	ja	.error_heap_corrupt
	jb	.error_bad_pointer

	;if newsize fits into real size of block, just reset size
	mov	ecx, [newsize]
	cmp	ecx, [ebx + HEAPBLOCK.size]
	ja	.current_block_too_small
	mov	[ebx + HEAPBLOCK.asize], ecx
	jmp	.done

.current_block_too_small:

	;ECX = requested size of block aligned
	add	ecx, HEAP_BLOCK_ALIGN-1
	and	ecx, not (HEAP_BLOCK_ALIGN-1)

	;EDX = required additional size
	mov	edx, ecx
	sub	edx, [ebx + HEAPBLOCK.size]

	;if there is no following block, we must allocate
	cmp	[ebx + HEAPBLOCK.last], 1
	je	.allocate_new_block
	ja	.error_heap_corrupt

	;ESI = following block
	lea	esi, [ebx + sizeof.HEAPBLOCK]
	add	esi, [ebx + HEAPBLOCK.size]

	;check following block
	cmp	[esi + HEAPBLOCK.sign], 'HB'
	jne	.error_heap_corrupt
	cmp	[esi + HEAPBLOCK.prev], ebx
	jne	.error_heap_corrupt

	;if following block is used, we must allocate
	cmp	[esi + HEAPBLOCK.used], 1
	ja	.error_heap_corrupt
	je	.allocate_new_block

	;if following block is not big enough, we must allocate
	mov	eax, [esi + HEAPBLOCK.size]
	add	eax, sizeof.HEAPBLOCK
	cmp	edx, eax
	ja	.allocate_new_block

.append_following_block:

	;EBX = our block, ESI=following block, EAX=size of new block (header included)
	add	[ebx + HEAPBLOCK.size], eax
	mov	eax, [newsize]
	mov	[ebx + HEAPBLOCK.asize], eax
	mov	[esi + HEAPBLOCK.sign], 0 ;delete mark

	;if next block was last, set this one as last
	mov	al, [esi + HEAPBLOCK.last]
	mov	[ebx + HEAPBLOCK.last], al
	cmp	al, 1
	ja	.error_heap_corrupt
	je	.block_created

	;otherwise point nextblock.prev to ours block
	add	esi, [esi + HEAPBLOCK.size] ;move to next block
	add	esi, sizeof.HEAPBLOCK
	mov	[esi + HEAPBLOCK.prev], ebx

	jmp	.block_created

.allocate_new_block:

	;allocate new block
	mov	ecx, [newsize]
	libcall mem.alloc, ecx
	jc	.rc

	;copy contents of old block into new block
	mov	edi, eax
	lea	esi, [ebx+sizeof.HEAPBLOCK]
	libcall mem.copy, edi, esi, [ebx+HEAPBLOCK.asize]
	jc	.rc

	;free old block
	libcall mem.free, esi
	jc	.rc

	;return new block
	mov	eax, edi
	jmp	.rnc

.block_created:

	;now check whether to split unused part from created block

	;ECX = aligned real size of block
	mov	ecx, [ebx + HEAPBLOCK.asize]
	add	ecx, HEAP_BLOCK_ALIGN-1
	and	ecx, not (HEAP_BLOCK_ALIGN-1)

	;EDX = gained space
	mov	edx, [ebx + HEAPBLOCK.size]
	sub	edx, ecx

	;if it's at least sizeof.HEAPBLOCK + HEAP_BLOCK_ALIGN, create new block
	sub	edx, sizeof.HEAPBLOCK
	jb	.done
	cmp	edx, HEAP_BLOCK_ALIGN
	jb	.done

	;ESI = new block
	mov	[ebx + HEAPBLOCK.size], ecx
	lea	esi, [ebx + ecx + sizeof.HEAPBLOCK]
	mov	[esi + HEAPBLOCK.sign], 'HB'
	mov	[esi + HEAPBLOCK.used], 0
	mov	[esi + HEAPBLOCK.size], edx
	mov	[esi + HEAPBLOCK.prev], ebx
	mov	al, [ebx + HEAPBLOCK.last]
	mov	[ebx + HEAPBLOCK.last], 0
	mov	[esi + HEAPBLOCK.last], al

	;if esi isn't last, set esi->next->prev = esi
	cmp	al, 1
	ja	.error_heap_corrupt
	je	@f
	lea	eax, [esi + edx + sizeof.HEAPBLOCK]
	mov	[eax + HEAPBLOCK.prev], esi
	@@:

.done:
	;return pointer
	lea	eax, [ebx + sizeof.HEAPBLOCK]

.rnc:	clc
.r:	pop	edi esi edx ecx ebx
	ret
.rc:	stc
	ret

.error_not_initialized:
	mov	eax, ERR_MODULE_NOT_INITIALIZED
	jmp	.rc

.error_heap_corrupt:
	mov	[mem.heap_corrupt], 1
	mov	eax, ERR_HEAP_CORRUPT
	jmp	.rc

.error_newsize_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_newsize_too_big:
	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc

.error_bad_pointer:
	mov	eax, ERR_INVALID_POINTER
	jmp	.rc
endp