;; @name mem.find
;; @desc
;;    Finds first occurence of given block of memory in another block of memory.
;;    Uses fast Horspool algorithm.
;; @arg block
;;    Pointer to block in which we will search.
;; @arg blocksize
;;    Size of block in which we search.
;; @arg pattern
;;    Pointer to pattern which we are looking for.
;; @arg patternsize
;;    Size of pattern block.
;; @ret
;;    CF = 1 on error, otherwise
;;    ZF = 1 if not found, otherwise
;;    EAX = offset (not pointer) in block, where the pattern was found
;; @err ERR_INVALID_POINTER
;;    some of pointers point to inaccessible area
;; @err ERR_ZERO_SIZE
;;    blocksize=0 or patternsize=0
;; @note Returned EAX can be in range 0 to block+blocksize-patternsize
;; @note Memory management doesn't have to be initalized for this to work
proc mem.find  block, blocksize, pattern, patternsize
local shift_table rb 256*4

	push	esi edi ecx edx ebx
	pushf
	cld

	;error if patternsize = 0
	cmp	[patternsize], 0
	je	.error_patternsize_zero

	;if patternsize>blocksize return nothing found
	mov	eax, [patternsize]
	cmp	eax, [blocksize]
	ja	.not_found


	;fill every shift_table member with patternsize
	libcall mem.fill, addr shift_table, 1024, [patternsize]
	jc	.rc

	;fill table with characters from pattern
	mov	ecx, [patternsize]
	dec	ecx
	jz	@f			;if patternsize=1, we don't need to fill table anymore
	lea	edx, [shift_table]
	mov	esi, [pattern]
.fill_table:
	movzx	eax, byte [esi]
	shl	eax, 2
	mov	dword [edx+eax], ecx
	inc	esi
	loop	.fill_table
@@:

	;go with ESI in block through blocksize-patternsize characters
	mov	esi, [block]		;esi = block
	mov	edi, [pattern]		;edi = pattern
	mov	ecx, esi
	add	ecx, [blocksize]
	sub	ecx, [patternsize]	;ecx = place where to stop searching
	lea	ebx, [shift_table]	;ebx = shift_table
.loop:

.check_match:
	;compare last character, and save it in DL
	mov	eax, [patternsize]
	mov	dl, byte [esi+eax-1]
	cmp	dl, byte [edi+eax-1]
	jne	.not_match

	;compare rest of pattern
	mov	eax, [patternsize]
	dec	eax
	libcall mem.cmp, esi, edi, [patternsize]
	jc	.r
	je	.found

	;we didn't find match, shift forward by number from shift_table[dl]
.not_match:
	movzx	eax, dl
	add	esi, [ebx + 4*eax]

	;continue loop
	cmp	esi, ecx
	jbe	.loop

	;nothing found, return ZF
.not_found:
	or	byte [esp], 1 shl 6 ; set ZF
	mov	eax, -1 ;for case someone forgets to check ZF
	jmp	.rnc

	;pattern found - return it's index in EAX, and ZF=0
.found: and	byte [esp], not (1 shl 6) ;clear ZF
	mov	eax, esi
	sub	eax, [block]

.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	ebx edx ecx edi esi
	ret 

.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_patternsize_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp

;thanks to mrpink for help with Horspool algo