;============================================================================
;; @name mem.realloc0
;; @desc
;;    Rellocates (resizes) block of memory.
;;    Block can be moved during this process, so you need to update pointer(s) to it.
;;    Contents of block are preserved.
;;    If block is enlarging, contents of added space are filled with zeroes.
;; @arg blockptr
;;    Pointer to block to relocate.
;;    This must be same pointer as returned from allocation function, pointer into block is not enough.
;; @arg newsize
;;    New size of block, in bytes
;; @ret
;;    CF set on error, orherwise
;;    EAX = pointer to reallocated memory block
;; @err ERR_INVALID_POINTER
;;    Blockptr is not a valid allocated block pointer.
;;    This could also mean {heap:heap corruption}, these two cases cannot be properly distinguished.
;; @err ERR_OUT_OF_MEM
;;     Not enough memory
;; @err ERR_ZERO_SIZE
;;     size = 0
;; @err ERR_OUT_OF_RANGE
;;     size is negative (or larger than 7FFFFFFFh if viewed as unsigned)
;; @err ERR_HEAP_CORRUPT
;;     {heap:Heap is corrupted}. You was writing to memory outside of allocated block.
;; @err ERR_MODULE_NOT_INITIALIZED
;;      Module is not initialized. Initialize module with {mem.init} to use this procedure.
;; @err ERR_UNKNOWN
;;     System returned error that wasn't translated to FASMLIB error.
;;     You can {contact} author about this, and translation can be added.
;; @dep mem.realloc_heap
;;     Only on platforms where FASMLIB heap manager is used.
;; @warn
;;     Block can be moved during realocation! If you had any pointers to block,
;;     they become invalid (so rather prefer using indexes to allocated blocks).
;;     If you saved block pointer in variable, you need to rewrite it with
;;     returned value.
;; @note Returned pointer is always >= 10000h to make it easily distinguishable from handles.
;; @note Maximal allowed newsize is 7FFFFFFFh, to catch errors with negative values.
;; @note Altough returned block may have bigger size than requested on some implementations,
;;       don't use this additional memory. {mem.realloc} doesn't preserve this memory.
;; @note Memory management must be {mem.init:initialized} for this procedure to work.
;; @note This is platform-specific procedure
proc mem.realloc0 blockptr, newsize
	push	ebx esi

	;error if newsize=0
	cmp	[newsize], 0
	je	.error_newsize_zero

	;ESI = save old block size
	libcall mem.size, [blockptr]
	mov	esi, eax

	;reallocate
	libcall mem.realloc, [blockptr], [newsize]
	jc	.r
	mov	ebx, eax

	;number of bytes to zero = newsize-oldsize
	mov	eax, [newsize]
	sub	eax, esi

	;zero new memory
	libcall mem.fill, addr ebx+esi, eax, 0
	jc	.r

	mov	eax, ebx

.rnc:	clc
.r:	pop	esi ebx
	ret
.rc:	stc
	jmp	.r

.error_newsize_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

endp