;; @name str.cmp
;; @desc
;;    Compares two strings, whether they are identical (case sensitive).
;;    If strings are different, tells which is greater (by comparing first different character).
;;    This is useful for sorting strings.
;; @arg str1
;;    Pointer to first string.
;; @arg str1len
;;    Size of buffer of first string.
;;    You can set this to -1 (FFFFFFFFh) if string is quaranteed to be properly zero-terminated.
;; @arg str2
;;    Pointer to second string.
;; @arg str2len
;;    Size of buffer of second string.
;;    You can set this to -1 (FFFFFFFFh) if string is quaranteed to be properly zero-terminated.
;; @ret
;;    CF set on error, otherwise
;;    {flagscmp:flags set for signed conditional jumps}, and
;;    if ZF=0, then EAX = offset where strings differ
;; @err ERR_ZERO_SIZE
;;    str1len=0 or str2len=0
;; @warn Use returned value in EAX with caution.
;;       It can point to end-of-string zero character. (like for "ab" and "abc").
;;       It can even point behind end of buffer, when string wasn't properly zero-terminated within
;;       it's buffer (like for "ab" in 2 byte buffer, and "abc").
;; @note If one string is longer than another, and shorter string is same as beginning of longer
proc str.cmp str1, str1len, str2, str2len
	push	ebx ecx edx esi edi
	pushf

	;error if str1len = 0
	cmp	[str1len], 0
	je	.error_str1len_zero

	;error if str2len = 0
	cmp	[str2len], 0
	je	.error_str2len_zero

	;EDI=str1, ESI=str2, EBX=str1len, ECX=str2len
	mov	edi, [str1]
	mov	ebx, [str1len]
	mov	esi, [str2]
	mov	ecx, [str2len]

	;EDX = index of current character = 0
	xor	edx, edx

	;scan each character
.loop:
	;check for end of buffers
	cmp	edx, [str1len]
	je	.str1_end
	cmp	edx, [str2len]
	je	.greater

	;compare characters
	mov	 al, [esi+edx]
	mov	 ah, [edi+edx]
	cmp	 ah, al
	ja	 .greater
	jb	 .lesser

	;if both are zero, then strings are equal
	cmp	 al, 0	 ;we know ah=al already
	je	 .equal

	inc	 edx
	jmp	 .loop


.str1_end:
	;we need to check if both buffers end aren't reached at same time
	cmp	edx, [str2len]
	je	.equal
	jmp	.lesser



	;flags for conditional jumps
	; je:      ZF=1
	; jne:     ZF=0
	; jg/jnle: ZF=0 & (SF=OF)
	; jge/jnl: SF=OF
	; jl/jnge: SF!=OF
	; jle/jng: ZF=1 | (SF!=OF)
.lesser:
	;ZF=0, SF=0, OF=1
	and	word [esp], not ((1 shl 6) + (1 shl 7)) ;clear ZF and SF
	or	word [esp], 1 shl 11			;set OF
	jmp	.done

.greater:
	;ZF=0, SF=1, OF=1
	and	word [esp], not (1 shl 6)		;clear ZF
	or	word [esp], (1 shl 11) + (1 shl 7)	;set OF and SF
	jmp	.done

.equal:
	;ZF=1, SF=1, OF=1
	or	word [esp], (1 shl 6) + (1 shl 7) + (1 shl 11)

.done:	mov	eax, edx
	and	byte [esp],not 1 ;clear CF
.r:	popf
	pop	edi esi edx ecx ebx
	ret

.rc:	or	byte [esp], 1 ;set CF
	jmp	.r

.error_str1len_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_str2len_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc
endp