;; @name stream.read
;; @desc
;;    Reads data from stream to buffer.
;;    Blocks until enough data is read, or EOF occurs.
;;    Returns OF, if EOF occurs before requested ammount of bytes is read.
;; @arg stream
;;    Handle of stream from which to read.
;; @arg buffer
;;    Pointer to buffer where data will be stored.
;; @arg count
;;    Number of bytes to read. Cannot be 0.
;; @arg readbuf
;;    Pointer to procedure that does actual reading. Procedure
;;    recieves pointer to memory where to read, and number of bytes to read.
;; @ret
;;    CF set on error, otherwise
;;    OF set if EOF was reached, and
;;    EAX = number of bytes readen.
;;    If OF=0, then EAX=count.
;; @err ERR_INVALID_HANDLE
;;    Invalid stream handle.
;; @err ERR_STREAM_NOT_READABLE
;;    Stream is not readable.
;; @err ERR_ZERO_SIZE
;;    count=0.
;; @err ERR_HANDLE_CORRUPT
;;    Stream is corrupt.
;;    Either you are overwriting stream object somewhere, or it is an error in FASMLIB.
;; @dep {stream read method}
proc stream.read, stream, buffer, count

	;error if count is zero
	cmp	[count], 0
	je	.error_count_zero

	;EAX = stream virtual table
	mov	eax, [stream]
	mov	eax, [eax + STREAM.vtab]

	;check if this is proper stream
	cmp	[eax + STREAM_VTAB.sign], 'STRM'
	jne	.error_not_stream

	;EAX = read method, error if not present
	mov	eax, [eax + STREAM_VTAB.read]
	cmp	eax, 0
	je	.error_stream_not_readable

	;call method
	libcall eax, [stream], [buffer], [count]
	jc	.rc

.rnc:	clc
.r:	ret
.rc:	stc
	jmp	.r

.error_not_stream:
	mov	eax, ERR_INVALID_HANDLE
	jmp	.rc

.error_count_zero:
	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_stream_not_readable:
	mov	eax, ERR_HANDLE_NOT_READABLE
	jmp	.rc

endp