;; @name text.extread.chars
;; @desc
;;     Reads from stream to buffer all characters which are in given
;;     character list.
;;     This is extended version of procedure, for simpler version see {text.read.chars}.
;; @arg stream
;;     Stream to read characters from.
;; @arg buffer
;;     Pointer to destination buffer, where the characters will be stored.
;; @arg buflen
;;     Length of destination buffer.
;; @arg charlist
;;     Null-terminated string containing list of characters to read.
;; @arg maxlen
;;     Maximum number of characters to read.
;;     Ignored if 0.
;; @arg len
;;     Pointer to dword variable, which on return holds number of characters readen.
;;     This is same value as one returned in EAX, and is only useful on ERR_BUFFER_FULL.
;;     Ignored if 0.
;; @ret
;;    CF=1 on error, otherwise
;;    EAX = number of characters readen, and
;;    ZF=1 if EAX=0 (no characters readen).
;; @note
;;    This function never reads character 0 from stream.
;; @err ERR_ZERO_SIZE
;;    buflen=0.
;;    len is set to 0.
;; @err ERR_BUFFER_FULL
;;    Destiantion buffer is full.
;;    Only buflen characters readen, and string in destination buffer is
;;    NOT zero-terminated.
;;    len is set to buflen.
;; @err ERR_HANDLE_EOF
;;    EOF reached without reading anything. len is set to 0.
proc text.extread.chars stream, buffer, buflen, charlist, maxlen, len
local peeklen dd ?
	push	ebx ecx edx edi
	pushf

	;error if length of buffer is 0
	cmp	[buflen], 0
	je	.error_buflen_zero

	;EDX = maximum number of chars allowed to skip (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;ECX = number of chars skipped
	xor	ecx, ecx

	;EDI = destination buffer
	mov	edi, [buffer]

	;EBX = peek 1 char
	;if we are on EOF before reading first char, return ERR_HANDLE_EOF
	libcall stream.peek.byte, [stream]
	jc	.rc
	jo	.error_eof
	jmp	.parse_char

.read_char:

	;EBX = peek 1 char
	libcall stream.peek.byte, [stream]
	jc	.rc
	jo	.done

.parse_char:
	movzx	ebx, al

	;stop reading at zero character
	test	al, al
	jz	.done

	;find out if current char is in charlist
	libcall str.poschar, [charlist], -1, ebx
	jc	.rc

	;if no, skipping is done
	jz	.done

	;store character
	mov	[edi], bl

	;skip char from stream
	libcall stream.skip, [stream], 1
	jc	.rc
	jo	.error_internal

	;another char done
	inc	ecx
	inc	edi

	;error if we filled entire buffer
	cmp	ecx, [buflen]
	je	.error_buffer_full
	ja	.error_internal

	;end loop if we skipped enough characters
	cmp	ecx, edx
	je	.done
	ja	.error_internal

	;continue loop
	jmp	.read_char

.done:
	;store ending 0 to buffer
	mov	byte [edi], 0

	;store number of readen chars to len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;return ECX
	mov	eax, ecx

	;if we skipped 0 characters, return ZF=1
	cmp	ecx, 0
	je	.rz

	;otherwise return ZF=0

.rnz:	and	byte [esp], not (1 shl 6) ;clear ZF
.rnc:	and	byte [esp], not 1 ;clear CF
.r:	popf
	pop	edi edx ecx ebx
	ret
.rc:	or	byte [esp], 1 ;set CF
	jmp	.r
.rz:	or	byte [esp], (1 shl 6) ;set ZF
	jmp	.rnc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

.error_buffer_full:
	;store number of readen chars to len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_BUFFER_FULL
	jmp	.rc

.error_buflen_zero:
	;store 0 to len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_ZERO_SIZE
	jmp	.rc

.error_eof:
	mov	eax, ERR_HANDLE_EOF
	jmp	.rc

endp