;; @name text.extread.dec_iq
;; @desc
;;    Reads signed decimal qword from stream.
;;    This is extended version, for simpler usage see {text.read.dec_iq}.
;;    Reads decimal digit characters ('0'-'9') until it reaches non-digit, EOF,
;;    or until resulting value overflows.
;; @arg stream
;;    Handle of stream from which to read.
;; @arg maxlen
;;    Maximum number of characters to read, including sign.
;;    0 if not specified.
;; @arg skipbl
;;    If 1, procedure first skips initial blank characters (space, tab, EOL).
;;    These characters don't count into maxlen size.
;;    These characters do not remain on stream in case of error.
;; @arg allowplus
;;    If 1, plus sign "+" is accepted.
;; @arg len
;;    Pointer to dword variable, that will be set to number of digits read.
;;    Ignored if 0.
;;    When skipbl=1, skipped initial blank characters don't count.
;; @ret
;;    CF set on error, otherwise
;;    EDX:EAX = readen qword value.
;; @err ERR_HANDLE_EOF
;;    EOF reached. len is set to 0.
;; @err ERR_OUT_OF_RANGE
;;    value < -9223372036854775808 or value > 9223372036854775807.
;;    In this case sign character and all digit characters remain on stream.
;;    len contains number of characters readen until overflow.
;; @err ERR_NOT_DECIMAL
;;    First character (after sign) is not a decimal digit.
;;    Characters remains on stream.
;;    len is set to 0.
;; @err ERR_ZERO_VALUE
;;    maxlen=0.
;;    len is set to 0.
;; @err ERR_INVALID_MODE
;;    skipbl > 1.
;;    len is set to 0.
;; @note Stream must have a read buffer.
;; @note If stream has static read buffer, it must be big enough to hold all characters
;;       of number.
proc text.extread.dec_iq stream, maxlen, skipbl, allowplus, len
local buflen dd ?
local sign dd ?
	push	ebx ecx esi edi

	mov	[sign], 0

	;skip blank chars
	;smartass thingy, but without it it would confuse people
	cmp	[skipbl], 1
	ja	.error_invalid_mode
	jne	@f
	libcall text.skip.chars, [stream], text.blank_chars
	jc	.rc
	@@:

	;EDX = maximum number of chars allowed to read  (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;ECX = number of char converted
	xor	ecx, ecx

	;peek at first char
	libcall stream.peek.byte, [stream]
	jc	.rc
	jo	.error_eof

	;check "+" (if allowplus=1)
	cmp	[allowplus], 1
	ja	.error_invalid_mode
	jne	@f
	cmp	al, '+'
	je	.get_another_char
	@@:

	;check '-'
	cmp	al, '-'
	jne	.check_first_digit
	mov	[sign], 1

	;there was a sign, peek another character
.get_another_char:
	libcall stream.peek, [stream], 2, addr buflen
	jc	.rc
	jo	.error_eof
	mov	al, byte [eax+1]
	inc	ecx

	;first character (after optional sign) must be digit
.check_first_digit:

	;error if it isn't dec char
	cmp	al, '0'
	jb	.error_not_numeric
	cmp	al, '9'
	ja	.error_not_numeric

	;EDI:EBX = result
	xor	edi, edi
	xor	ebx, ebx

.load_char:

	;peek 1 more char
	lea	eax, [ecx+1]
	libcall stream.peek, [stream], eax, addr buflen
	jc	.rc
	jo	.done
	mov	esi, eax

.convert_char:
	movzx	eax, byte [esi + ecx]

	;end if it isn't hex digit
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	ja	.done

	;convert to hex digit
	sub	al, '0'

	;multiply value with 10
	push	eax edx esi
	xor	edx, edx
	mov	esi, 10
	mov	eax, ebx
	mul	esi
	mov	ebx, eax
	mov	eax, edi
	mov	edi, edx
	mul	esi
	add	eax, edi
	mov	edi, eax
	pop	esi edx eax

	;add digit
	add	ebx, eax
	adc	edi, 0
	jc	.error_overflow

	;check if we have read enough chars
	inc	ecx
	cmp	ecx, edx
	ja	.error_internal
	je	.done

	;if we don't have another char buffered, read it. otherwise continue converting next char
	cmp	ecx, [buflen]
	ja	.error_internal
	jb	.convert_char
	jmp	.load_char

.done:
	;skip readen chars
	libcall stream.skip, [stream], ecx
	jc	.rc
	jo	.error_internal

	;save number of chars readen to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;set sign and check overflow
	cmp	[sign], 0
	jne	.negative
.positive:
	cmp	edi, 0x7FFFFFFF
	ja	untaken .error_overflow
	jmp	.sign_set
.negative:
	cmp	edi, 0x80000000
	jb	taken .negative_okay
	ja	untaken .error_overflow
	cmp	ebx, 0
	jne	untaken .error_overflow
	jmp	.sign_set  ;number is exactly 0x8000000000000000
.negative_okay:
	not	ebx
	not	edi
	add	ebx, 1
	adc	edi, 0
.sign_set:

	;store result
	mov	edx, edi
	mov	eax, ebx

	clc
.r:	pop	edi esi ecx ebx
	ret
.rc:	stc
	jmp	.r


.error_invalid_mode:
	;save 0 to variable pointed by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_INVALID_MODE
	jmp	.rc

.error_eof:
	;save 0 to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_HANDLE_EOF
	jmp	.rc

.error_not_numeric:
	;save 0 to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_NOT_DECIMAL
	jmp	.rc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

.error_len_zero:
	;save 0 to variable pointet by len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_ZERO_VALUE
	jmp	.rc

.error_overflow:
	;save ecx to variable pointet by len
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp