;; @name text.extread.hex_q
;; @desc
;;    Reads unsigned hexadecimal qword from stream.
;;    This is extended version, for simpler usage see {text.read.hex_d}.
;;    Reads hexadecimal digit characters ('0'-'9','a'-'f','A'-'F') until it reaches non-digit, EOF,
;;    or until resulting value overflows.
;; @arg stream
;;    Handle of stream from which to read.
;; @arg maxlen
;;    Maximum number of digits to read.
;;    0 if not specified.
;; @arg skipbl
;;    If 1, procedure first skips initial blank characters (space, tab, EOL).
;;    These characters don't count into maxlen size.
;; @arg len
;;    Pointer to dword variable, that will be set to number of digits read.
;;    Ignored if 0.
;;    When skipbl=1, skipped initial blank characters don't count.
;; @ret
;;    CF set on error, otherwise
;;    EDX:EAX = readen qword
;; @err ERR_HANDLE_EOF
;;    EOF reached. len is set to 0.
;; @err ERR_OUT_OF_RANGE
;;    value > FFFFFFFFFFFFFFFF.
;;    In this case all digit characters remain on stream.
;;    len contains number of characters readen until overflow.
;; @err ERR_NOT_HEXADECIMAL
;;    First character is not hexadecimal digit.
;;    Character remains on stream.
;;    len is set to 0.
;; @err ERR_ZERO_VALUE
;;    maxlen=0.
;;    len is set to 0.
;; @err ERR_INVALID_MODE
;;    skipbl > 1.
;;    len is set to 0.
;; @note Stream must have a read buffer.
;; @note If stream has static read buffer, it must be big enough to hold all digits
;;       of number.
proc text.extread.hex_q stream, maxlen, skipbl, len
local buflen dd ?
	push	ebx ecx esi edi

	;skip blank chars
	;smartass thingy, but without it it would confuse people
	cmp	[skipbl], 1
	ja	.error_invalid_mode
	jne	@f
	libcall text.skip.chars, [stream], text.blank_chars
	jc	.rc
	@@:

	;EDX = maximum number of chars allowed to read  (0 if not specified)
	mov	edx, -1
	mov	eax, [maxlen]
	cmp	eax, 0
	cmovne	edx, eax

	;error if maximal number of chars to read is zero
	cmp	edx, 0
	je	.error_len_zero

	;peek at first char
	libcall stream.peek.byte, [stream]
	jc	.rc
	jo	.error_eof

	;error if it isn't hex char
	cmp	al, '0'
	jb	.error_not_numeric
	cmp	al, '9'
	jbe	@f
	cmp	al, 'A'
	jb	.error_not_numeric
	cmp	al, 'F'
	jbe	@f
	cmp	al, 'a'
	jb	.error_not_numeric
	cmp	al, 'f'
	ja	.error_not_numeric
	@@:

	;ECX = number of char converted
	xor	ecx, ecx

	;EDI:EBX = result
	xor	edi, edi
	xor	ebx, ebx

.load_char:

	;peek 1 more char
	lea	eax, [ecx+1]
	libcall stream.peek, [stream], eax, addr buflen
	jc	.rc
	jo	.done
	mov	esi, eax

.convert_char:
	mov	al, [esi + ecx]

	;end if it isn't hex digit
	cmp	al, '0'
	jb	.done
	cmp	al, '9'
	jbe	.numeric
	cmp	al, 'A'
	jb	.done
	cmp	al, 'F'
	jbe	.upcase_letter
	cmp	al, 'a'
	jb	.done
	cmp	al, 'f'
	ja	.done

.lowcase_letter:
	sub	al, 'a'-10
	jmp	.add_digit

.upcase_letter:
	sub	al, 'A'-10
	jmp	.add_digit

.numeric:
	sub	al, '0'

.add_digit:
	;shift to left by 4 (one hex digit)
	test	edi, 0xF0000000
	jnz	.error_overflow
	shld	edi, ebx, 4
	shl	ebx, 4

	;add digit
	movzx	eax, al
	add	ebx, eax
	adc	edx, 0
	jc	.error_overflow

	;check if we have read enough chars
	inc	ecx
	cmp	ecx, edx
	ja	.error_internal
	je	.done

	;if we don't have another char buffered, read it. otherwise continue converting next char
	cmp	ecx, [buflen]
	ja	.error_internal
	jb	.convert_char
	jmp	.load_char

.done:
	;skip readen chars
	mov	eax, [buflen]
	libcall stream.skip, [stream], eax
	jc	.rc
	jo	.error_internal

	;save number of chars readen to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	;return result
	mov	edx, edi
	mov	eax, ebx

	clc
.r:	pop	edi esi ecx ebx
	ret
.rc:	stc
	jmp	.r

.error_invalid_mode:
	;save 0 to variable pointed by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_INVALID_MODE
	jmp	.rc

.error_eof:
	;save 0 to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_HANDLE_EOF
	jmp	.rc

.error_not_numeric:
	;save 0 to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	dword [eax], 0
	@@:

	mov	eax, ERR_NOT_HEXADECIMAL
	jmp	.rc

.error_internal:
	mov	eax, ERR_FASMLIB
	jmp	.rc

.error_len_zero:
	mov	eax, ERR_ZERO_VALUE
	jmp	.rc

.error_overflow:
	;save ecx to variable pointet by len, if desired
	mov	eax, [len]
	cmp	eax, 0
	je	@f
	mov	[eax], ecx
	@@:

	mov	eax, ERR_OUT_OF_RANGE
	jmp	.rc
endp